"use strict"

module.exports = Registry

var DEFAULT_TTL = 60 // seconds
var Etcd = require("etcdjs")

var debug = require("util").debuglog("nsolid")

function Registry(cluster) {
  if (!(this instanceof Registry)) {
    return new Registry(cluster)
  }
  if (typeof cluster != "string") {
    throw new Error("please specify an Etcd address")
  }
  this.etcd = new Etcd(cluster, {json: true})
}

Registry.prototype.register = function (procInfo, ttl, callback) {
  // register as present w/ ttl
  if (!procInfo.app) {
    throw new Error("`app` required for process registration")
  }
  if (!procInfo.id) {
    throw new Error("`id` required for process registration")
  }
  if (!procInfo.address) {
    throw new Error("`address` required for process registration")
  }

  if (typeof ttl == "function") {
    callback = ttl
    ttl = DEFAULT_TTL
  }

  ttl = ttl >>> 0
  if (callback == null) {
    callback = function noop() {}
  }

  var key = "/nsolid/" + procInfo.app + "/" + procInfo.id
  debug("registering with etcd ", key)
  this.etcd.set(key, procInfo, {ttl: ttl}, callback)
}

function flatten(nodes) {
  var addresses = []
  nodes.forEach(function eachNode(node) {
    if (node.value) {
      addresses.push(node.value)
    }
    if (node.nodes) {
      addresses = addresses.concat(flatten(node.nodes))
    }
  })
  return addresses
}

// returns array of objects: [{app: app, id: id, address: address, ...}]
Registry.prototype.list = function list(app, callback) {
  if (callback == null) {
    throw new Error("Must provide a callback")
  }
  var wildcard = false
  if (app == "*") {
    app = ""
    wildcard = true
  }
  var key = "/nsolid/" + app
  function parseResults(err, reply) {
    debug("parseResults for list")
    if (err) {
      return callback(err)
    }
    if (reply == null || reply.node == null || reply.node.nodes == null) {
      return callback(err, [])
    }
    return callback(err, flatten(reply.node.nodes))
  }
  debug("fetching list results from etcd")
  this.etcd.get(key, {recursive: true}, parseResults)
}

// TODO Registry.prototype.filter = function filter(app, idPrefix, callback) {

// returns object: {app: app, id: id, address: address, ...}
Registry.prototype.lookup = function lookup(app, id, callback) {
  if (app == null || id == null) {
    throw new Error("`app` and `id` are both required for lookup")
  }
  if (callback == null) {
    throw new Error("Must provide a callback")
  }
  var key = "/nsolid/" + app + "/" + id
  function parseResults(err, reply) {
    debug("parseResults for lookup")
    if (err) {
      return callback(err)
    }
    if (reply == null) {
      return callback(err, null)
    }
    return callback(err, reply.node.value)
  }
  debug("fetching key from etcd ", key)
  this.etcd.get(key, parseResults)
}

// Registry.prototype.watch = function watch(app) {
//   // watch for any changes at the app level (new nodes, node expiry)
//   // optionally pass sequence to make sure nothing is missed
// }
