var test = require("tape")

var Registry = require("../registry")
var Etcd = require("etcdjs")

test("init", function (t) {
  var registry = new Registry("localhost:4001")
  t.ok(registry instanceof Registry, "yay constructors")
  t.ok(registry.etcd instanceof Etcd, "has an etcd")
  t.end()
})

test("register", function (t) {
  var registry = new Registry("localhost:4001")
  registry.register({app: "test-foo", id: 12345, address: "localhost:12000"}, function cb(err, reply) {
    t.notOk(err, "no error")
    t.ok(reply.action, "reply has an action (etcd-ish)")
    t.equal(reply.action, "set", "we set a value")
    t.ok(reply.node, "reply has a node")
    t.equal(reply.node.ttl, 60, "default ttl is 60 for register")
    t.end()
  })
})

test("list with no matches", function (t) {
  var registry = new Registry("localhost:4001")
  registry.list("test-???", function (err, nodes) {
    t.notOk(err, "no error")
    t.deepEqual(nodes, [], "no nodes matched")
    t.end()
  })
})

test("list", function (t) {
  var registry = new Registry("localhost:4001")
  var count = 0
  function cb() {
    count++
    next()
  }
  registry.register({app: "test-zap", id: 12345, address: "localhost:12000"}, cb)
  registry.register({app: "test-zap", id: "abc12", address: "localhost:12001"}, cb)
  registry.register({app: "test-bar", id: "a1b2", address: "localhost:12002"}, cb)
  registry.register({app: "test-bar", id: 12345, address: "localhost:12003"}, cb)

  function next() {
    if (count < 4) {
      return
    }
    registry.list("test-bar", function (err, nodes) {
      t.notOk(err, "no error")
      t.equals(nodes.length, 2, "only got nodes from test-bar")
      t.ok(nodes[0].app && nodes[0].id && nodes[0].address, "has expected fields")
      t.ok(nodes[1].app && nodes[1].id && nodes[1].address, "has expected fields")
      t.equals(nodes[0].app, "test-bar", "app matches")
      t.equals(nodes[1].app, "test-bar", "app matches")
      t.end()
    })
  }
})

test("list all apps", function (t) {
  var registry = new Registry("localhost:4001")
  var count = 0
  function cb() {
    count++
    next()
  }
  registry.register({app: "test-zap", id: 12345, address: "localhost:12000"}, cb)
  registry.register({app: "test-zap", id: "abc12", address: "localhost:12001"}, cb)
  registry.register({app: "test-bar", id: "a1b2", address: "localhost:12002"}, cb)
  registry.register({app: "test-bar", id: 12345, address: "localhost:12003"}, cb)

  function next() {
    if (count < 4) {
      return
    }
    registry.list("*", function (err, nodes) {
      t.notOk(err, "no error")
      t.ok(nodes.length >= 4, "got all of them")
      t.ok(nodes[0].app && nodes[0].id && nodes[0].address, "has expected fields")
      t.ok(nodes[1].app && nodes[1].id && nodes[1].address, "has expected fields")
      t.ok(nodes[2].app && nodes[2].id && nodes[2].address, "has expected fields")
      t.ok(nodes[3].app && nodes[3].id && nodes[3].address, "has expected fields")
      t.end()
    })
  }
})

test("list after expire", function (t) {
  var registry = new Registry("localhost:4001")
  var count = 0

  registry.register({app: "test-yyy", id: 12345, address: "localhost:12000"}, 1, onRegister)

  function onRegister() {
    registry.list("test-yyy", function (err, nodes) {
      t.notOk(err, "no error")
      t.equals(nodes.length, 1, "not expired yet")
      t.ok(nodes[0].app && nodes[0].id && nodes[0].address, "has expected fields")
      t.equals(nodes[0].app, "test-yyy", "app matches")
      setTimeout(afterExpire, 1500)
    })
  }

  function afterExpire() {
    registry.list("test-yyy", function (err, nodes) {
      t.notOk(err, "no error")
      t.equals(nodes.length, 0, "no nodes left")
      t.end()
    })
  }

})

test("lookup with no matches", function (t) {
  var registry = new Registry("localhost:4001")
  registry.lookup("test-???", "a1b2", function (err, node) {
    t.notOk(err, "no error")
    t.equal(node, null, "no node found")
    t.end()
  })
})

test("lookup", function (t) {
  var registry = new Registry("localhost:4001")
  var count = 0
  function cb() {
    count++
    next()
  }
  registry.register({app: "test-zzz", id: 12345, address: "localhost:12000"}, cb)
  registry.register({app: "test-zzz", id: "a1b1", address: "localhost:12001"}, cb)
  registry.register({app: "test-aaa", id: 12345, address: "localhost:12003"}, cb)

  function next() {
    if (count < 3) {
      return
    }
    registry.lookup("test-zzz", 12345, function (err, node) {
      t.notOk(err, "no error")
      t.ok(node.app && node.id && node.address, "has expected fields")
      t.equals(node.app, "test-zzz", "app matches")
      t.equals(node.id, 12345, "app matches")
      t.equals(node.address, "localhost:12000", "app matches")
      t.end()
    })
  }

})
