/**
 * Server-side lib for collect
 */

var async = require('async')
var createRouter = require('routes')
var url = require('url')
var camelize = require('camelize')
var EventEmitter = require('events').EventEmitter
var path = require('path')
var fs = require('fs')
var mkdirp = require('mkdirp')
var spawn = require('child_process').spawn
var intersect = require('./collect/intersect')

var buildResponse = require('./collect/build-response')
var hub = require('./collect/hub')
var st = require('st')
var skateboard = require('skateboard')
var computeAppId = require('./compute-app-id')

// webpack dev/hot middleware configuration
var developmentMode = process.env.NODE_ENV !== 'production'

var NSOLID_CLI = process.env.NSOLID_CLI || 'nsolid-cli'

var commands = [
  'system_info', // hostname, totalmem, and metadata
  'system_stats',
  'process_stats',
  'info'
]

// TODO: read these from disk
var snapshots = {}
var profiles = {}

function broadcastToHub (cmd, cb) {
  hub(NSOLID_CLI, cmd, cb)
}

function poll (pollInterval, duration, broadcast) {
  var ee = new EventEmitter()

  function run () {
    async.mapSeries(commands, broadcast || broadcastToHub, function (err, commandResults) {

      if (err) {
        setTimeout(run, pollInterval)
        return console.error(err)
      }

      var resultsByCommand = {}

      commands.forEach(function (name, i) {
        resultsByCommand[name] = clean(commandResults[i])
      })

      ee.emit('input', resultsByCommand)

      var processes = intersect(resultsByCommand['info'], resultsByCommand['process_stats'])
      var intersectedHosts = intersect(resultsByCommand['system_info'], resultsByCommand['system_stats'])

      var hostsByHostname = {}
      var hosts = {}

      Object.keys(intersectedHosts).forEach(function (processId) {
        var host = intersectedHosts[processId].reply

        if (!hostsByHostname[host.hostname]) {
          hostsByHostname[host.hostname] = host
        }

        // TODO: we could do better if we normalize hosts and add
        //       a ref to a top-level 'hosts' object inside of processes
        //       and applications
        hosts[processId] = hostsByHostname[host.hostname]
      })

      var array = buildResponse(processes, hosts, pollInterval, duration)

      array.sort(function (a, b) {
        return a.name.localeCompare(b.name)
      })

      array.forEach(function (app) {
        app.snapshots = snapshots[app.name] || []
        app.profiles = profiles[app.name] || []

        app.processes.sort(function (a, b) {
          return a.pid - b.pid
        })

        app.processes.forEach(function (process) {
          process.snapshots = app.snapshots.filter(function (snapshot) {
            var same = snapshot.id === process.id
            if (same) {
              snapshot.pid = process.pid
              snapshot.title = process.title
            }

            return same
          })

          process.profiles = app.profiles.filter(function (profile) {
            var same = profile.id === process.id
            if (same) {
              profile.pid = process.pid
              profile.title = process.title
            }

            return same
          })
        })
      })

      ee.emit('output', array)

      // for testing
      if (pollInterval !== false) {
        setTimeout(run, pollInterval)
      }
    })
  }

  // kick it off
  run()

  return ee
}

function clean (a) {
  // Until we can land a fix for https://github.com/nodesource/nsolid-epm-cli/issues/2
  if (!a) {
    return []
  }

  return a.filter(function (o) {
    // TODO: it's likely better to show the errors on the client if something
    //       bad is happening.  In all of the cases I've seen so far it's just
    //       an inability to resolve an ipv6 address (derp)
    return !!o.reply
  }).map(function (o) {
    o.reply = camelize(o.reply)
    return o
  })
}

function bailOnError (message, fn, context) {
  return function () {
    console.error(message)
    try {
      return fn.apply(context, arguments)
    } catch(e) {
      console.log(e.stack)
    }
  }
}

function spawnNsolidCli (cmd, params) {
  params.timeout = 1000

  var args = Object.keys(params).map(function (key) {
    return '--' + key + '=' + params[key]
  })

  args.push(cmd)

  return spawn(NSOLID_CLI, args, {
    stdio: 'pipe',
    env: process.env
  })
}

// TODO: refactor - this argument list is way too long
function respond (storagePath, storageDirectory, extension, res, params, cmd, store) {
  var c = spawnNsolidCli(cmd, {
    app: params.app,
    id: params.id
  })
  c.stderr.pipe(process.stderr)

  var time = Date.now()
  var filename = generateFilename(params.app, params.id, time, extension)

  c.stdout.pipe(fs.createWriteStream(path.join(storagePath, storageDirectory, filename)))

  c.on('exit', function (code) {
    if (!store[params.app]) {
      store[params.app] = []
    }

    var insert = {
      time: time,
      app: params.app,
      id: params.id,
      date: new Date(),
      file: filename
    }

    // put the latest snapshots at the beginning of the array
    store[params.app].unshift(insert)

    res.setHeader('content-type', 'application/json')
    res.end(JSON.stringify(insert) + '\n')
  })
}

function generateFilename (appName, processId, time, extension) {
  var appId = computeAppId(appName)
  return [appId, processId, time].join('-') + '.' + extension
}

function createServerRoutes (storagePath) {
  var router = createRouter()

  router.addRoute(
    '/heapsnapshot',
    function createHeapSnapshot (req, res, match) {
      var method = req.method.toLowerCase()
      var parsedUrl = url.parse(req.url, true)
      var filename
      var params = parsedUrl.query

      if (method === 'get') {
        filename = generateFilename(params.app, params.id, parsedUrl.query.time, 'heapsnapshot')
        var readStream = fs.createReadStream(path.resolve(storagePath, path.join('snapshots', filename)))
        readStream.on('open', function () {

          res.setHeader('Content-Type', 'application/json')
          res.setHeader('Content-Disposition', 'attachment; filename=' + filename)

          readStream.pipe(res)
        })
        readStream.on('error', function (error) {
          res.end(error.toString())
        })
      } else if (method === 'post') {
        var time = Date.now()
        filename = generateFilename(params.app, params.id, time, 'heapsnapshot')

        var c = spawnNsolidCli('snapshot', params)
        c.stderr.pipe(process.stderr)

        c.stdout.pipe(fs.createWriteStream(path.resolve(storagePath, path.join('snapshots', filename))))

        c.on('exit', function (code) {
          if (!snapshots[params.app]) {
            snapshots[params.app] = []
          }

          var insert = {
            time: time,
            app: params.app,
            id: params.id,
            date: new Date(),
            file: filename
          }

          // put the latest snapshots at the beginning of the array
          snapshots[params.app].unshift(insert)

          res.setHeader('content-type', 'application/json')
          res.end(JSON.stringify(insert) + '\n')
        })
      } else {
        res.end()
      }
    }
  )

  router.addRoute(
    '/profile',
    function createProfile (req, res, match) {
      var method = req.method.toLowerCase()
      var parsedUrl = url.parse(req.url, true)
      var params = parsedUrl.query

      if (method === 'get') {
        // e.g.: http://localhost:3000/applications/default-nsolid-scenario/processes/1062671c412debb5fdd538e08a333bd1f51dd152/profile?time=1438049368666
        var filename = generateFilename(params.app, params.id, parsedUrl.query.time, 'cpuprofile')
        var readStream = fs.createReadStream(path.resolve(storagePath, path.join('profiles', filename)))
        readStream.on('open', function () {

          res.setHeader('Content-Type', 'application/json')
          res.setHeader('Content-Disposition', 'attachment; filename=' + filename)

          readStream.pipe(res)
        })

        readStream.on('error', function (error) {
          if (error.code === 'ENOENT') {
            res.writeHead(404)
          } else {
            res.writeHead(500)
          }
          res.end(error.toString())
        })
      } else if (method === 'post') {
        var interval = 5
        if (params.interval) {
          var intermediary = parseInt(params.interval, 10)
          if (!isNaN(intermediary)) {
            interval = intermediary
          }
        }

        // curl -X POST :3000/applications/default-nsolid-scenario/processes/1062671c412debb5fdd538e08a333bd1f51dd152/profile
        var c = spawnNsolidCli('profile_start', {
          app: params.app,
          id: params.id
        })
        c.on('exit', function (code) {
          setTimeout(function () {
            respond(storagePath, 'profiles', 'cpuprofile', res, params, 'profile_stop', profiles)
          }, interval * 1000)
        })
      } else {
        res.end()
      }
    }
  )

  return router
}

function createServer (port, iface, storagePath, pollingEmitter, listening, error) {
  if (developmentMode) {
    var webpackConfig = require('../webpack.config')
    var webpack = require('webpack')
    var compiler = webpack(webpackConfig)
    var webpackDevMiddleware = require('webpack-dev-middleware')(compiler, webpackConfig.middleware)
    var webpackHotMiddleware = require('webpack-hot-middleware')(compiler, webpackConfig.hotMiddleware)
  }

  function processMiddleware (req, res, fn) {
    // only run the middleware when developmentMode is enabled
    if (developmentMode) {
      webpackDevMiddleware(req, res, function (err) {
        if (err) return fn(err)
        webpackHotMiddleware(req, res, fn)
      })
    } else {
      fn()
    }
  }

  var workerPath = path.join(__dirname, 'collect', 'heap-snapshot', 'worker.js')
  var router
  var applicationClients = []

  if (storagePath) {
    var snapshotPath = path.join(storagePath, 'snapshots')
    var profilePath = path.join(storagePath, 'profiles')
    try {
      mkdirp.sync(snapshotPath)
      mkdirp.sync(profilePath)
    } catch (err) {
      return error(err)
    }
    router = createServerRoutes(storagePath)
  }

  var mount = st({
    path: path.join(__dirname, '..', developmentMode ? 'assets' : 'build'),
    index: developmentMode ? false : 'index.html',
    passthrough: true
  })

  skateboard({
    port: port,
    interface: iface,
    dir: path.join(__dirname, '..', developmentMode ? 'assets' : 'build'),
    requestHandler: handleRequest,
    staticHandler: mount,
    listening: listening,
    error: error
  }, function (stream, params) {
    if (params.source === 'applications') {
      applicationClients.push(stream)
      lastSend && stream.write(lastSend)
      stream.on('end', function () {
        applicationClients = applicationClients.filter(function (a) {
          return a !== stream
        })
      })
    } else if (params.source === 'heapsnapshot') {

      var proc = spawn('node', [workerPath, '--dir=' + snapshotPath], {
        stdio: 'pipe'
      })

      stream.once('end', bailOnError('client disconnected from heapsnapshot worker', proc.kill, proc))
      proc.on('exit', bailOnError('heapsnapshot worker exited', stream.close, stream))
      proc.stderr.pipe(process.stderr)
      stream.pipe(proc.stdin)

      proc.stdout.setEncoding('utf8')
      proc.stdout.pipe(stream)
    }
  })

  function handleRequest (req, res) {
    processMiddleware(req, res, function () {
      var urlPath = url.parse(req.url).pathname
      var match = router && router.match(urlPath)
      if (match) { // serve dynamic routes such as profiles and heap snapshots
        match.fn(req, res, match)
      } else { // serve static files
        mount(req, res, function () {
          req.url = '/'
          // `st` has cached the url as a 404, so override the cache
          req.sturl = '/'
          processMiddleware(req, res, function () {
            mount(req, res, function () {
              res.writeHead(404, 'Not Found')
              res.end('Not Found')
            })
          })
        })
      }
    })
  }

  var lastSend = null
  pollingEmitter.on('output', function (applications) {
    lastSend = JSON.stringify(applications)
    applicationClients.forEach(function (client) {
      client.write(lastSend)
    })
  })
}

// INTERFACE
module.exports.poll = poll
module.exports.createServer = createServer
