var merge = require('merge');

module.exports = computeSparklines

var graphs = {}
function createGraph(name, storeTime, pollInterval) {
  if (!graphs[name]) {
    graphs[name] = {}
  }

  var graph = graphs[name]

  return {
    add: function addRecord(name, value, longestGraph) {

      if (!graph[name]) {
        graph[name] = []
      }
      var now = Date.now();

      var minDate = now - storeTime;

      var subgraph = graph[name]
      // TODO: handle the case where the previous record is > 2*interval
      subgraph.push({ v: value, t: now })

      var l = subgraph.length
      var i = 0

      // Instead of relying on pollInterval, find the first result that is
      // out of the allowed time range.  Using that index we can slice off
      // collected metrics before it and maintain a consistent array length.

      for (; i < l; i++) {
        if (subgraph[i].t >= minDate) {
          break;
        }
      }

      if (i) {
        var maxLength = l - i
        graph[name] = graph[name].slice(-(maxLength))
      } else {
        var diff = longestGraph.length - subgraph.length;

        if (diff > 0) {
          for (var j=0; j<diff; j++) {
            // since we're unshifting we need to add the longestGraph
            // times in reverse order. hence the `diff - j` oddness
            subgraph.unshift({
              v: null,
              t: longestGraph[diff - j].t
            })
          }
        }
      }
    },
    toJSON: function () {
      return graph
    }
  }
}

function getMaxGraphLength () {
  var graphKeys = Object.keys(graphs);
  var l = graphKeys.length
  var maxLength = 0
  var maxGraph

  for (var i=0; i<l; i++) {
    var graph = graphs[graphKeys[i]]
    var subGraphKeys = Object.keys(graph)
    var k = subGraphKeys.length
    for (var j=0; j<k; j++) {
      var subgraph = graph[subGraphKeys[j]]
      if (subgraph.length > maxLength) {
        maxGraph = subgraph
        maxLength = maxLength
      }
    }
  }

  return maxGraph || [];
}

function computeSparklines(application, pollInterval, duration) {
  var storeTime = (duration * 1000 * 60)
  var longestGraph = getMaxGraphLength();

  var appGraph = createGraph('app-' + application.name, storeTime, pollInterval)

  var totals = {
    rss: 0,
    total: 0,
    used: 0
  }

  application.processes.forEach(function (proc) {
    var processGraph = createGraph('process-' + proc.id, storeTime, pollInterval)

    // sum data across all processes
    totals.rss += proc.memory.rss
    totals.total += proc.memory.total
    totals.used += proc.memory.used

    processGraph.add('rss', proc.memory.rss, longestGraph)
    processGraph.add('heapTotal', proc.memory.total, longestGraph)
    processGraph.add('heapUsed', proc.memory.used, longestGraph)
    processGraph.add('asyncHandles', proc.async.handles, longestGraph)
    processGraph.add('asyncRequests', proc.async.requests, longestGraph)

    proc.graphs = processGraph.toJSON()

    if (!proc.host.graphs) {
      var host = proc.host
      var hostGraph = createGraph('host-' + host.hostname, storeTime, pollInterval)

      hostGraph.add('load1m', host.load1m, longestGraph)
      hostGraph.add('memoryPercentageUsed', (host.totalmem - host.freemem) / host.totalmem, longestGraph);
      host.graphs = hostGraph.toJSON()
    }
  })

  // add new app data
  var l = application.processes.length

  appGraph.add('rss', totals.rss / l, longestGraph)
  appGraph.add('heapTotal', totals.total / l, longestGraph)
  appGraph.add('heapUsed', totals.used / l, longestGraph)

  application.graphs = appGraph.toJSON()
}
