"use strict"

module.exports = Listener

var inherits = require("util").inherits
var EventEmitter = require("events").EventEmitter
var http = require("http")
var url = require("url")
var concat = require("./concat")
var os = require("os")
var decode = require("querystring").decode

var debug = require("util").debuglog("nsolid")

function Listener(socket) {
  if (!(this instanceof Listener)) {
    return new Listener(socket)
  }
  EventEmitter.call(this)

  // port
  // ip:port
  // ip
  // ip:?
  // ip:
  var listenOptions = {}
  if (isNumber(socket) || socket == null) {
    listenOptions.port = socket
  }
  else if (typeof socket == "string" && socket.match(/:/)) {
    var ips = externalInterfaces()
    if (ips.indexOf(socket) >= 0) {
      // No port and matches a known IP address
      listenOptions.host = socket
    }
    else {
      // Look for a port specified
      var parts = socket.split(":")
      var port = parts.pop()
      // might be IPv6
      var host = parts.join(":")

      if (ips.indexOf(host) >= 0) {
        // known ip + port
        listenOptions.host = host
        listenOptions.port = (port == null || port == "" || port == "?") ? 0 : port
      }
      // Fall-through for unknown (private ip or hostname that will be DNS looked up)
      else if (port == null || port == "") {
        // IPv6 with an empty final set, e.g. :: (localhost)
        listenOptions.port = 0
        if (parts.length > 1) {
          // IPv6 with empty final set, e.g. ::
          parts.push("")
        }
      }
      else if (port == "?") {
        // An explicit random port, e.g. :::? or 127.0.0.1:?
        listenOptions.port = 0
      }
      else {
        // A port number, e.g. 127.0.0.1:8008 or :::8008
        listenOptions.port = port
      }
      // Re-assemble the host, accounting for IPv6
      listenOptions.host = parts.join(":")
    }
  }
  else {
    listenOptions.host = socket
    listenOptions.port = 0
  }

  debug("nsolid listener starting up with listenOptions", listenOptions)

  var self = this
  var handler = function handler(req, res) {
    var replySent = false
    debug("nsolid agent incoming request")
    // closure to finish reply
    function next(err, content) {
      if (replySent) {
        debug("repeated call to handler! (ignoring)")
        return
      }
      replySent = true
      // send content back...
      if (err != null) {
        console.error("NSolid Agent Listener error", err)
        res.statusCode = err.status || 503
        return res.end(JSON.stringify(err.message))
      }
      if (content == null) {
        debug("agent listener sending empty reply")
        return res.end()
      }

      // stream duck-typing
      if (content.pipe != null && (typeof content.pipe == "function")) {
        debug("piping stream reply in agent listener")
        content.pipe(res)
        content.on("error", function (err) {
          console.error("NSolid Agent Listener stream content error")
          console.error(err.message)
          console.error(err.stack)
          res.statusCode = err.status || 503
          return res.end(JSON.stringify(err.message))
        })
        return
      }

      if (content instanceof Error) {
        debug("nsolid agent listener sending error reply")
        content = content.toString()
      }
      debug("nsolid agent listener sending single-chunk reply")
      return res.end(JSON.stringify(content))
    }

    // parse route into rpc
    var rpcSpec = self.parse(req)
    req.pipe(concat(function (body) {
      if (body.length === 0) {
        rpcSpec.data = null
      }
      else {
        debug("nsolid agent listener request contains a body")
        try {
          rpcSpec.data = JSON.parse(body.toString())
        } catch (e) {
          console.error("NSolid Agent Listener unable to parse request body, aborting request")
          return next(new Error("Failed to parse request body"))
        }
      }
      var hadListener = self.emit(rpcSpec.command, req, rpcSpec, next)
      if (!hadListener) {
        // Allow a "wildcard" command intercept
        var hadWildcard = self.emit("*", req, rpcSpec, next)
        if (!hadWildcard) {
          debug("nsolid agent listener no handler installed for %s", rpcSpec.command)
          var err = new Error("NSolid Agent: no command handler installed for `" + rpcSpec.command + "`")
          err.status = 404
          return next(err)
        }
        debug("nsolid listener dispatched wildcard rpc (not expected for agent)")
      }
    }))
  }

  this.server = http.createServer(handler)
  this.server.listen(listenOptions, function () {
    self.emit("listening")
  })
}
inherits(Listener, EventEmitter)

// "rest-ish" RPC
// v1: /command or /command/id
// v2: /command or /command/id OR /command?id=id or /command?app=appname
// v3: v2 + timeout option
Listener.prototype.parse = function parse(req) {
  var rpc = {
    command: "",
    data: null,
    id: null,
    app: null,
    timeout: null
  }
  debug("nsolid agent incoming request: %s", req.url)
  var parsed = url.parse(req.url)
  var parts = parsed.pathname.split("/")
  var params = decode(parsed.query)
  rpc.command = parts[1].toLowerCase()
  if (rpc.command == "*") {
    // prevent manual wildcard
    rpc.command = ""
  }
  if (parts[2] != null && parts[2] != "") {
    rpc.id = parts[2]
  }
  if (params.id != null && params.id != "") {
    rpc.id = params.id
  }
  if (params.app != null && params.app != "") {
    rpc.app = params.app
  }
  if (params.timeout != null) {
    if (+params.timeout == params.timeout) {
      rpc.timeout = +params.timeout
    }
  }
  return rpc
}

function isNumber(n) {
  return !isNaN(parseFloat(n)) && isFinite(n)
}

function externalInterfaces() {
  var interfaces = os.networkInterfaces()
  var devices = Object.keys(interfaces)
  var addresses = []
  // Collect all external addresses
  devices.forEach(function eachDevice(device) {
    interfaces[device].forEach(function eachAddr(addr) {
      if (!addr.internal) {
        addresses.push(addr.address)
      }
    })
  })
  return addresses
}

Listener.prototype.getAddress = function getAddress() {
  var address = this.server.address()
  if (typeof address == "string") {
    // Unix socket
    // NOTE: for now, the rpc client does not support unix sockets
    //   here for safety but not expected
    debug("nsolid agent listener bound to unix socket (unexpected)")
    return {port: "", ips: [address]}
  }

  debug("nsolid listener listening on ", address)

  var ip = address.address
  if (ip == "::" || ip == "0.0.0.0") {
    // Looks like all interfaces are bound
    var addresses = externalInterfaces()

    debug("nsolid agent listener sending %s ip options", addresses.length)
    return {port: address.port, ips: addresses}
  }
  else {
    debug("nsolid agent listener bound to specific ip %s", ip)
    return {port: address.port, ips: [ip]}
  }
}
