var test = require("tape")

var Client = require("../rpcclient")

var http = require("http")
var concat = require("terminus").concat
var spigot = require("stream-spigot")
var url = require("url")

test("simple", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route is as expected")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    c.send("ping", null).pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("simple error", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route is as expected")
    res.statusCode = 503
    res.end("NOT OK")
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    var req = c.send("ping", null)
    req.on("error", function (err) {
      t.ok(err instanceof Error, "err is an Error")
      t.equal(err.message, "NOT OK", "converted 503 response to error")
      t.end()
    })
    req.pipe(concat(function (body) {
      t.fail("Shouldn't be here")
    }))
  })
  s.unref()
})

test("data", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/foo", "route is as expected")
    req.pipe(concat(function (body) {
      t.equal(body.toString(), JSON.stringify("BAR"), "jsonified body matches")
      res.end("OK")
    }))
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    c.send("foo", "BAR").pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("object data", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/foo", "route is as expected")
    req.pipe(concat(function (body) {
      t.equal(body.toString(), JSON.stringify({aaa: "BBB"}), "jsonified body matches")
      res.end("OK")
    }))
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    c.send("foo", {aaa: "BBB"}).pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("routeTo id", function (t) {
  var s = http.createServer(function (req, res) {
    var parsed = url.parse(req.url)
    t.equal(parsed.pathname, "/ping", "route has command")
    t.equal(parsed.query, "id=abc")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    var c = new Client(hubUri)
    c.routeTo("ping", {id: "abc"}, null).pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("routeTo id w/ data", function (t) {
  var s = http.createServer(function (req, res) {
    var parsed = url.parse(req.url)
    t.equal(parsed.pathname, "/ping", "route has command")
    t.equal(parsed.query, "id=abc")
    req.pipe(concat(function (body) {
      t.equal(body.toString(), JSON.stringify("BAR"), "jsonified body matches")
      res.end("OK")
    }))
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    var c = new Client(hubUri)
    c.routeTo("ping", {id: "abc"}, "BAR").pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("routeTo app", function (t) {
  var s = http.createServer(function (req, res) {
    var parsed = url.parse(req.url)
    t.equal(parsed.pathname, "/ping", "route has command")
    t.equal(parsed.query, "app=abc")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    c.routeTo("ping", {app: "abc"}, null).pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("routeToBuffered id", function (t) {
  var s = http.createServer(function (req, res) {
    var parsed = url.parse(req.url)
    t.equal(parsed.pathname, "/ping", "route has command")
    t.equal(parsed.query, "id=abc")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    var c = new Client(hubUri)
    c.routeToBuffered("ping", {id: "abc"}, null, function (err, res, body) {
      t.notOk(err, "no error")
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    })
  })
  s.unref()
})

test("routeToBuffered timeout", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route is as expected")
    setTimeout(function () {
      res.end("NOPE")
    }, 2000)
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    var c = new Client(hubUri)
    c.routeToBuffered("ping", null, null, function (err, res, body) {
      t.ok(err, "have an error")
      t.notOk(body, "No body for an error")
      t.end()
    }, 1000)
  })
  s.unref()
})

test("routeToBuffered app", function (t) {
  var s = http.createServer(function (req, res) {
    var parsed = url.parse(req.url)
    t.equal(parsed.pathname, "/ping", "route has command")
    t.equal(parsed.query, "app=abc")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    var c = new Client(hubUri)
    c.routeToBuffered("ping", {app: "abc"}, null, function (err, res, body) {
      t.notOk(err, "no error")
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    })
  })
  s.unref()
})

test("sendTo", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route includes id")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    Client.sendTo(hubUri, "ping", null).pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("sendTo data", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/foo", "route is as expected")
    req.pipe(concat(function (body) {
      t.equal(body.toString(), JSON.stringify("BAR"), "jsonified body matches")
      res.end("OK")
    }))
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    Client.sendTo(hubUri, "foo", "BAR").pipe(concat(function (body) {
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    }))
  })
  s.unref()
})

test("sendToBuffered", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route includes id")
    res.end("OK")
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    Client.sendToBuffered(hubUri, "ping", null, function (err, response, body) {
      t.notOk(err, "no error")
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    })
  })
  s.unref()
})

test("sendToBuffered data", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/foo", "route is as expected")
    req.pipe(concat(function (body) {
      t.equal(body.toString(), JSON.stringify("BAR"), "jsonified body matches")
      res.end("OK")
    }))
  })
  s.listen(function () {
    var hubUri = "localhost:" + this.address().port
    Client.sendToBuffered(hubUri, "foo", "BAR", function (err, response, body) {
      t.notOk(err, "no error")
      t.equal(body.toString(), "OK", "server response as expected")
      t.end()
    })
  })
  s.unref()
})

test("hub down", function (t) {
  var hubUri = "http://localhost:44444"
  var c = new Client(hubUri)
  var r = c.send("ping", null)
  r.pipe(concat(function (body) {
    t.fail("Should have an error")
    t.end()
  }))
  r.on("error", function (err) {
    t.ok(err, "an error, as expected")
    t.end()
  })
})

test("remote error", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route is as expected")
    req.socket.destroy()
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    var r = c.send("ping", null)
    r.pipe(concat(function (body) {
      t.fail("Should have an error")
      t.end()
    }))
    r.on("error", function (err) {
      t.ok(err, "an error, as expected")
      t.end()
    })
  })
  s.unref()
})

test("custom timeout (test takes 1 second)", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route is as expected")
    setTimeout(function () {
      res.end("NOPE")
    }, 2000)
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    var r = c.send("ping", null, 1000)
    r.pipe(concat(function (body) {
      t.fail("Should have timed out")
      t.end()
    }))
    r.on("error", function (err) {
      t.ok(err, "have an error")
      t.end()
    })
  })
  s.unref()
})

test("default timeout (test takes 10 seconds)", function (t) {
  var s = http.createServer(function (req, res) {
    t.equal(req.url, "/ping", "route is as expected")
    var timer = setTimeout(function () {
      res.end("NOPE")
    }, 11000)
    timer.unref()
  })
  s.listen(function () {
    var hubUri = "http://localhost:" + this.address().port
    var c = new Client(hubUri)
    var r = c.send("ping", null)
    r.pipe(concat(function (body) {
      t.fail("Should have timed out")
      t.end()
    }))
    r.on("error", function (err) {
      t.ok(err, "have an error")
      t.end()
    })
  })
  s.unref()
})
